type 't env = (string * 't) list

let rec lookup x y = match x with
                         | (i1,e1) :: x1 -> if y = i1 then e1 else lookup x1 y
                         | [] -> failwith("wrong env")

type typ =
  | TypI                                (* int                         *)
  | TypB                                (* bool                        *)
  | TypF of typ * typ                   (* (argumenttype, resulttype)  *)

(* abstract syntax with explicit types *)

type tyexpr =
  | CstI of int
  | CstB of bool
  | Var of string
  | Let of string * tyexpr * tyexpr
  | Prim of string * tyexpr * tyexpr
  | If of tyexpr * tyexpr * tyexpr
  | Letfun of string * string * typ * tyexpr * typ * tyexpr
          (* (f,       x,       xTyp, fBody,  rTyp, letBody *)
  | Call of tyexpr * tyexpr



(* Type checking *)

let rec typ (e : tyexpr) (env : typ env) : typ =
    match e with
    | CstI i -> TypI
    | CstB b -> TypB
    | Var x  -> lookup env x
    | Prim(ope, e1, e2) ->
      let t1 = typ e1 env in
      let t2 = typ e2 env in
      (match (ope, t1, t2) with
      | ("*", TypI, TypI) -> TypI
      | ("+", TypI, TypI) -> TypI
      | ("-", TypI, TypI) -> TypI
      | ("=", TypI, TypI) -> TypB
      | ("<", TypI, TypI) -> TypB
      | ("&", TypB, TypB) -> TypB
      | _   -> failwith "unknown op, or type error")
    | Let(x, eRhs, letBody) ->
      let xTyp = typ eRhs env in
      let letBodyEnv = (x, xTyp) :: env in
      typ letBody letBodyEnv
    | If(e1, e2, e3) ->
      (match typ e1 env with
      | TypB -> let t2 = typ e2 env in
                let t3 = typ e3 env in
                if t2 = t3 then t2
                else failwith "If: branch types differ"
      | _    -> failwith "If: condition not boolean")
    | Letfun(f, x, xTyp, fBody, rTyp, letBody) ->
      let fTyp = TypF(xTyp, rTyp) in
      let fBodyEnv = (x, xTyp) :: (f, fTyp) :: env in
      let letBodyEnv = (f, fTyp) :: env in
      if typ fBody fBodyEnv = rTyp
      then typ letBody letBodyEnv
      else failwith ("Letfun: return type in " ^ f)
    | Call(Var f, eArg) ->
      (match lookup env f with
      | TypF(xTyp, rTyp) ->
        if typ eArg env = xTyp then rTyp
        else failwith "Call: wrong argument type"
      | _ -> failwith "Call: unknown function")
    | Call(_, eArg) -> failwith "Call: illegal function in call"

let typeCheck e = typ e [];;
